''' dAmnViper.data module
    Created by photofroggy.
    
    This module provides the Channel object, which represents a
    joined channel. The object stores information about the channel,
    such as the title/topic, and users in the channel.
'''

from dAmnViper.parse import Packet

class Channel(object):
    """ Objects representing dAmn channels.
    
        Information about the channels are stored in here. The different
        attributes of the object are as follows:
        
        * ``title`` - This is an instance of the class
          ``Channel.header``, and stores information about the title of
          the channel. The object has the attributes ``content``, ``by``
          and ``ts``. These store the content of the title, the person
          who set the title, and the timestamp from when they set it,
          respectively.
        * ``topic`` - Similar to the ``title`` attribute, but it stores
          information about the channel's topic, as opposed to the
          channel's title.
        * ``pc`` - Stores information about the channel's privclasses.
        * ``pc_order`` - A list storing the the privclass names in order
          of hierarchy. This is to allow applications to display
          information more easily if needed.
        * ``member`` - A dictionary storing information about each user
          currently in the channel.
        * ``namespace`` - The channel's full namespace. This is usually
          a string in the form ``chat:channel_name``. This can differ
          depending on the type of channel the object represents.
        * ``shorthand`` - The shorthand form of the channel
          ``namespace``. This is usually a string in the format
          ``#channel_name`` if the namespace is of the format
          ``chat:channel_name``. This can differ depending on the type
          of channel the object represents.
        * ``type`` - A string in for format ``<dAmn channel
          'namespace'>`` where ``namespace`` is the same as the object's
          ``namespace`` attribute.
        
        Calling ``str(channel)``, where ``channel`` is an instance of
        the ``Channel`` class, returns the ``namespace`` attribute.
        
        As an example, here is what some of the attributes would look
        like for the **#Botdom** channel when the object is created::
            
            >>> channel = Channel('chat:Botdom', '#Botdom')
            >>> channel.namespace
            'chat:Botdom'
            >>> channel.shorthand
            '#Botdom'
            >>> str(channel)
            'chat:Botdom'
        
        The other attributes would be empty objects of their respective
        types until the appropriate data is received from the server.
    """
    
    class header:
        def __init__(self):
            self.content = ''
            self.by = ''
            self.ts = 0.0
    
    def __init__(self, namespace, shorthand):
        """Set up all our variables."""
        self.title = Channel.header()
        self.topic = Channel.header()
        self.pc = {}
        self.pc_order = []
        self.member = {}
        
        self.namespace = namespace
        self.shorthand = shorthand
        
        self.type = '<dAmn channel \''+self.namespace+'\'>'
    
    def process_property(self, data):
        """ Called when receive a property packet for the channel.
            
            This method makes sure that the data is stored in the right
            places in the object.
        """
        if data['p'] == 'title':
            self.title.content = data['value']
            self.title.by = data['by']
            self.title.ts = data['ts']
        if data['p'] == 'topic':
            self.topic.content = data['value']
            self.topic.by = data['by']
            self.topic.ts = data['ts']
        if data['p'] == 'privclasses':
            self.pc = Packet(data['value'], ':').args
            self.pc_order = sorted(self.pc.keys(), key=int)
            self.pc_order.reverse()
        if data['p'] == 'members':
            member = Packet(data['value'])
            while member.cmd != None and len(member.args) > 0:
                self.register_user(member)
                member = Packet(member.body)
    
    def register_user(self, info, user = None):
        """ Called when a user joins the channel.
            
            Simply store their information in the ``member`` dictionary.
        """
        user = user if user != None else info.param
        if user in self.member:
            self.member[user]['con']+= 1
        else:
            self.member[user] = info.args
            self.member[user]['con'] = 1
    
    def __str__(self):
        return self.namespace

# EOF



''' dAmnViper.base module
    Created by photofroggy.
    
    This module provides the dAmnSock class, which acts
    as an API for connecting to and interacting with
    deviantART.com's chatrooms. This is achieved using
    Twisted.
'''

# Standard library
import sys
import time
from functools import wraps


# Twisted imports. Wooooo.
from twisted.internet import defer
from twisted.internet import reactor


# Viper imports
# Data
from dAmnViper.data import Channel
# Parsing
from dAmnViper.parse import Packet
from dAmnViper.parse import ProtocolParser
# Internets! lols.
from dAmnViper.net import ConnectionFactory


class Client(object):
    """ This class is a client for llama-like chat servers.
        
        The llama project can be seen here:
            http://code.google.com/p/project-llama-server/
            http://llamaserver.blogspot.com/
        
        Applications using this class should extend it and override the
        `startedConnecting` and `teardown` methods. If you are simply
        connecting to a dAmn server, you may want to use the
        :py:class:`dAmnClient class <dAmnViper.base.dAmnClient>` instead.
    """
    
    class platform:
        """ Information about the dAmn Viper platform. """
        name = 'dAmn Viper'
        version = 3
        state = 'Beta'
        build = 56
        stamp = '26072011-013703'
        series = 'Twister'
        author = 'photofroggy'
    
    
    class User(object):
        """ User login data. """
        def __init__(self):
            self.username = None
            self.cookie = None
            self.token = None
    
    
    class Flag(object):
        """ Status flags are stored here """
        
        debug = False
        
        def __init__(self):
            self.connecting = False
            self.shaking = False
            self.loggingin = False
            self.connected = False
            self.autorejoin = True
            self.disconnecting = False
            self.reconnect = False
            self.quitting = False
            self.retry = False
            self.restart = False
            self.close = False
    
    
    class Constants(object):
        """ dAmnSock style 'constants'. """
        def __init__(self):
            self.SERVER = None
            self.CLIENT = None
            self.PORT = None
    
    
    class session:
        status = (False, None)
        cookie = None
        token = None
    
    
    class Connection:
        def __init__(self):
            self.disconnects = 0
            self.attempts = 0
            self.limit = 3
    
    
    class Defer:
        """ Storage for Deferred and IDelayedCall objects. """
        
        def __init__(self):
            self.loop = None
            self.timeout = None
        
        def teardown(self):
            """ Stop any delayed calls from running. """
            if self.loop is None and self.timeout is None:
                return
            
            try:
                self.loop.cancel()
            except Exception as e:
                pass
            
            try:
                self.timeout.cancel()
            except Exception as e:
                pass
    
    
    extras = {'remember_me':'1'}
    agent = 'dAmnViper (python) dAmnSock/1.1'
    info = {}
    
    conn = None
    io = None
    
    Protocol = ProtocolParser
    
    autojoin = ['chat:Botdom']
    channel = {}
    
    timeout_delay = 120
    
    default_ns = '~Global'

    def __init__(self, *args, **kwargs):
        # Create an instance of our protocol class. Do anything else required.
        self.populate_objects()
        self.agent = 'dAmnViper/{0} (Python) viper/base/Client/{1}.{2}'.format(
            self.platform.series, self.platform.version, self.platform.build)
        self.default_ns = '~Global'
        self.init(*args, **kwargs)
    
    def init(self, *args, **kwargs):
        """ Overwrite this method if you need to do anything
            when an instance of this class is created.
        """
        pass
    
    def populate_objects(self):
        """ Populate our objects that are used to store information and stuff. """
        self.user = self.User()
        self.flag = self.Flag()
        self.CONST = self.Constants()
        self.connection = self.Connection()
        self.defer = self.Defer()
        self.protocol = self.Protocol()
    
    def nullflags(self):
        """ Reset all status flags in this client. """
        self.flag = self.Flag()
    
    def set_protocol(self, protocol=None):
        """ Store the given IO protocol. """
        self.io = protocol
        
        if protocol is None:
            self.persist()
        
    def start(self, *args, **kwargs):
        """ Start the client. """
        self.nullflags()
        
        self.connection.attempts = 1
        
        # Open a connection to the server.
        self.makeConnection()
        
        # Set up the client's main loop.
        self.defer.loop = reactor.callLater(1, self.mainloop, args, kwargs)
    
    def makeConnection(self):
        """ Opens a connection to a dAmn-like server.
            
            Raises a ``ValueError`` if ``Client.user.username`` or
            ``Client.user.token`` is ``None``.
            
            Raises a ``ValueError`` if any of the ``Client.CONST`` attributes
            are ``None``.
        """
        if None in (self.user.username, self.user.token):
            raise ValueError('.user.username and .user.token must be set.')
        
        if None in (self.CONST.SERVER, self.CONST.CLIENT, self.CONST.PORT):
            raise ValueError('The server, client and port to use must be defined.')
        
        write_pair = self.get_write_pair()
        
        # Make a connection.
        self.conn = ConnectionFactory(self, write_pair[0], write_pair[1])
        reactor.connectTCP(self.CONST.SERVER, self.CONST.PORT, self.conn)
    
    def startedConnecting(self, connector):
        """ This method is called when twisted is starting to connect.
            
            Sub classes of the ``Client`` class must override this method in
            order to determine what the client should do when trying to
            connect to the chat server.
            
            Raises a ``NotImplementedError`` if not overridden.
        """
        raise NotImplementedError
    
    def connectionLost(self, connector, reason):
        """ This method is called when we lose our connection. """
        self.logger('** Connection closed.{0}'.format(
            ' Reason: {0}'.format(reason) if isinstance(reason, str) else ''),
            showns=False)
        
        self.on_connection_lost(connector, reason)
        
        self.set_protocol(None)
    
    def connectionFailed(self, connector, reason):
        """ This method is called when we fail to connect. """
        self.logger('** Failed to connect.{0}'.format(
            ' Reason: {0}'.format(reason) if isinstance(reason, str) else ''),
            False)
        
        if self.connection.attempts == self.connection.limit:
            self.logger('** Failed to connect {0} times in a row.'.format(
                self.connection.attempts), showns=False)
        else:
            self.flag.retry = True
        
        self.on_connection_failed(connector, reason)
        
        self.set_protocol(None)
    
    def connectionMade(self):
        """ This method is called when we have made a connection. """
        if not self.flag.connecting:
            return
        
        self.on_connection_made()
        
        self.handshake()
    
    def persist(self):
        """ Determine what we should do when we have fully lost our connection
            to the server.
        """
        
        if not self.flag.quitting or self.flag.reconnect:
            self.connection.attempts = 1
            self.logger('** Attempting to reconnect...', showns=False)
            self.makeConnection()
            return
        
        if self.flag.retry:
            self.logger('** Attempting to connect again...', showns=False)
            self.connection.attempts+= 1
            self.makeConnection()
            return
        
        self.defer.teardown()
        try:
            self.teardown()
        except Exception as e:
            pass
        return
    
    def teardown(self):
        """ Override this method to do stuff when the client gives up.
            It is a good idea to use this method to call reactor.stop().
        """
        raise NotImplementedError
    
    def on_connection_start(self, connector):
        """ Override as needed. """
        pass
    
    def on_connection_lost(self, connector, reason):
        """ Override as needed. """
        pass
    
    def on_connection_failed(self, connector, reason):
        """ Override as needed. """
        pass
    
    def on_connection_made(self):
        """ Override as needed. """
        pass
    
    def mainloop(self, args, kwargs):
        """ This is the client's main loop. """
        self.on_loop(*args, **kwargs)
        self.defer.loop = reactor.callLater(1, self.mainloop, args, kwargs)
    
    def on_loop(self, *args, **kwargs):
        """ Overwrite this if you need to do anything on the main application loop. """
        pass
    
    def timedout(self):
        """ Timeout detected.
            
            If this method gets called, then the client has not received
            any data for 2 minutes. Send a pong to test the connection.
        """
        self.pong()
    
    def send(self, data):
        """ Send data to dAmn! """
        if self.io is None:
            return 0
        return self.io.send_packet(data)
    
    def close(self):
        """ This is how we close our connection! """
        self.set_protocol()
        
        if self.defer.timeout is not None:
            try:
                self.defer.timeout.cancel()
            except:
                pass
            self.defer.timeout = None
    
    def format_ns(self, ns):
        """ This takes a dAmn channel name and formats it as a raw
            dAmn namespace.
        """
        ns = str(ns)
        un = self.user.username
        pre = ns[:1]
        if pre == '#':
            return 'chat:{0}'.format(ns[1:])
        if pre == '@':
            if not un:
                return 'pchat:{0}'.format(ns[1:])
            para = [ns[1:], un]
            para.sort(key=str.lower)
            return 'pchat:{0}:{1}'.format(para[0], para[1])
        if ns[:5] == 'chat:' or ns[:6] == 'pchat:':
            return ns
        return 'chat:'+ns
    
    def deform_ns(self, ns):
        """ This does the opposite of format_ns() """
        parts = str(ns).split(':')
        discard = self.user.username
        if parts[0].lower() == 'chat':
            return '#{0}'.format(parts[1])
        if parts[0].lower() == 'pchat':
            if not discard:
                return '@{0}:{1}'.format(parts[1], parts[2])
            parts.pop(0)
            for i in range(0,2):
                if parts[i].lower() == discard.lower():
                    parts.pop(i)
                    return '@{0}'.format(parts[0])
        if ns[:1] in ('#', '@'):
            return ns
        return '#'+ns
    
    def handle_pkt(self, packet, stamp):
        """ Handle packets as they come in. """
        if self.defer.timeout is not None:
            self.defer.timeout.cancel()
        self.defer.timeout = reactor.callLater(self.timeout_delay, self.timedout)
        
        ns = '~Global'
        
        if packet.param:
            if packet.param[:5] in ('chat:', 'pchat'):
                ns = self.deform_ns(packet.param)
            elif packet.param[:6] == 'login:':
                ns = '@' + packet.param[6:]
        
        data = self.protocol.mapper(packet)
        loglist = self.protocol.logger(data['event'], data['rules'], ns, packet)
        
        if loglist is not None:
            self.logger(*loglist, ts=stamp)
        
        getattr(self, 'pkt_' + data['event'], self.pkt_unknown)(data['args'])
        self.pkt_generic(data)
        return data
    
    # PROTOCOL OUTPUT
    # The methods below pretty much define the protocol for outgoing packets.

    def raw(self, data):
        """ Send a raw dAmn packet. """
        return self.send(data)
    
    def handshake(self):
        """ Send a handshake to the server. """
        self.flag.shaking = True
        
        pkt = '{0}\nagent={1}\n{2}'.format(
            self.CONST.CLIENT, self.agent,
            '\n'.join(['{0}={1}'.format(
                    key, self.info[key]
            ) for key in self.info]))
        
        return self.send(pkt)
    
    def login(self):
        """ Send our login packet. Set the loggingin flag to true.
            
            Raises a ``ValueError`` if ``Client.user.username`` or
            ``Client.user.token`` is ``None``.
        """
        if None in (self.user.username, self.user.token):
            raise ValueError('.user.username and .user.token must be set.')
        
        self.flag.loggingin = True
        return self.send('login {0}\npk={1}\n'.format(self.user.username, self.user.token))
        
    def pong(self):
        """ Send a pong packet to the server """
        return self.send('pong\n')
        
    def join(self, ns):
        """ Send a join packet to dAmn. """
        return self.send('join {0}\n'.format(ns))
        
    def part(self, ns):
        """ Send a part packet to dAmn. """
        return self.send('part {0}\n'.format(ns))
        
    def say(self, ns, message):
        """ Send a message to a dAmn channel namespace. """
        return self.send('send {0}\n\nmsg main\n\n{1}'.format(ns, str(message)))
    
    def action(self, ns, message):
        """ Send an action to a dAmn channel namespace. """
        return self.send('send {0}\n\naction main\n\n{1}'.format(ns, str(message)))
    
    def me(self, ns, message):
        """ This is just another way to do an action. """
        return self.action(ns, message)
        
    def npmsg(self, ns, message):
        """ Send a non-parsed message to a dAmn channel namespace. """
        return self.send('send {0}\n\nnpmsg main\n\n{1}'.format(ns, str(message)))
    
    def promote(self, ns, user, pc=None):
        """ Promote a user in a dAmn channel. """
        return self.send('send {0}\n\npromote {1}\n{2}'.format(ns, user, '' if pc == None else '\n'+str(pc)))
    
    def demote(self, ns, user, pc=None):
        """ Demote a user in a dAmn channel. """
        return self.send('send {0}\n\ndemote {1}\n{2}'.format(ns, user, '' if pc == None else '\n'+str(pc)))
    
    def kick(self, ns, user, r=None):
        """ Kick a user out of a dAmn channel. """
        return self.send('kick {0}\nu={1}\n{2}'.format(ns, user, '' if r == None else '\n'+str(r)))
    
    def ban(self, ns, user):
        """ Ban a user from a dAmn channel. """
        return self.send('send {0}\n\nban {1}\n'.format(ns, user))
    
    def unban(self, ns, user):
        """ Unban someone from a dAmn channel. """
        return self.send('send {0}\n\nunban {1}\n'.format(ns, user))
    
    def get(self, ns, p):
        """ Get a property for a dAmn channel. """
        return self.send('get {0}\np={1}\n'.format(ns, p))
    
    def set(self, ns, p, val):
        """ Set a property for a dAmn channel. """
        return self.send('set {0}\np={1}\n\n{2}'.format(ns, p, val))
    
    def admin(self, ns, command):
        """ Send an admin command to a dAmn channel. No need for multiple methods here. """
        return self.send('send {0}\n\nadmin\n\n{1}'.format(ns, command))
    
    def disconnect(self):
        """ Send a disconnect packet to the dAmn server. """
        return self.send('disconnect\n')
        
    def kill(self, ns, r=None):
        """ Send a kill packet to the dAmn server. """
        return self.send('kill {0}\n{1}'.format(ns, '' if r == None else '\n'+str(r)))
    
    # END PROTOCOL OUTPUT METHODS
    # BEGIN PROTOCOL INPUT METHODS
    # These methods process incomming data.
    
    def pkt_generic(self, data):
        """ We have received a packet!
            
            Override this method to do stuff whenever a packet is
            received. If you want to do stuff when a specific packet is
            received, define a new ``pkt_*`` method for that packet.
            
            If the method you are defining already exists, then you
            should call the original method from your own definition, so
            that the client does not break.
            
            Here, the method is given a dict as returned by the
            :py:class:`ProtocolParser <dAmnViper.parse.ProtocolParser>`
            method ``mapper``. Other ``pkt_*`` methods are only given
            the data under the ``args`` key of the dict returned by the
            ``mapper`` method.
        """
        pass
    
    def pkt_unknown(self, data):
        """ We received something unexpected. Oh well. """
        pass
        
    def pkt_login(self, data):
        """ Received a login packet.
            
            When we receive a login packet, we need to determine whether
            the login was successful or not, and act based on that.
            
            This method closes the client if the login failed.
            Otherwise, the channels listed in the ``autojoin`` attribute
            are joined.
            
            This method is essential for connecting to dAmn properly.
        """
        
        self.flag.loggingin = False
        
        if data['e'] != 'ok':
            self.nullflags()
            self.close()
            return
        
        self.flag.loggingin = False
        self.flag.connecting = False
        self.flag.connected = True
        self.connection.attempts = 0
        
        for ns in self.autojoin:
            self.join(self.format_ns(ns))
    
    def pkt_join(self, data):
        """ Received a join packet.
            
            This is sent by the server when the client tries to join a
            channel on the server.
            
            If the join was successful, a :py:class:`Channel object
            <dAmnViper.data.Channel>` is created for the channel and
            stored in the ``channel`` attribute of the client.
            
            If the join failed, the client disconnects if there are no
            other joined channels. (``naive``)
        """
        if data['e'] == 'ok':
            self.channel[data['ns']] = Channel(
                data['ns'], self.deform_ns(data['ns'])
            )
            return
        
        if len(self.channel) > 0:
            return
        
        self.handle_pkt(Packet('disconnect\ne=no joined channels\n\n'), time.time())
    
    def pkt_part(self, data):
        """ Received a part packet.
            
            Similar to ``pkt_join``. This method determines whether or
            not the client is being kicked off the server.
        """
        if data['ns'] in self.channel.keys():
            del self.channel[data['ns']]
        
        if len(self.channel) > 0:
            return
        
        if 'r' in data.keys():
            if data['r'] in ('bad data', 'bad msg', 'msg too big') or 'killed:' in data['r']:
                self.handle_pkt(
                    Packet('disconnect{0}e={1}{2}{3}'.format("\n", data['r'], "\n", "\n")),
                    time.time())
                return
        
        if data['e'] != 'ok':
            return
        
        if self.channel or self.flag.disconnecting or self.flag.quitting:
            return
        
        self.handle_pkt(Packet('disconnect\ne=no joined channels\n\n'), time.time())
    
    def pkt_property(self, data):
        """ Received a channel property packet.
            
            This method makes sure that the information received is
            stored in the right place.
        """
        if data['p'] == 'info':
            return
        
        if not data['ns'] in self.channel.keys():
            return
        
        self.channel[data['ns']].process_property(data)
    
    def pkt_recv_join(self, data):
        """ Received a recv_join packet.
            
            This happens when a user joins a channel in which the client
            is also present.
            
            This method simply stores information about the user that
            just joined.
        """
        self.channel[data['ns']].register_user(Packet(data['info']), data['user'])
        
    def pkt_recv_part(self, data):
        """ Received a recv_part packet.
            
            This happens when a user leaves a channel in which the
            client is present.
            
            Here, we remove any records of the user being in the channel.
        """
        if not data['user'] in self.channel[data['ns']].member:
            return
        
        self.channel[data['ns']].member[data['user']]['con']-= 1
        
        if self.channel[data['ns']].member[data['user']]['con'] == 0:
            del self.channel[data['ns']].member[data['user']]
            
    def pkt_recv_kicked(self, data):
        """ Received a recv_kick packet.
            
            Similar to the ``pkt_recv_part`` method.
        """
        if not data['user'] in self.channel[data['ns']].member:
            return
        del self.channel[data['ns']].member[data['user']]
        
    def pkt_recv_privchg(self, data):
        """ Received a recv_privchg packet.
            
            This happens when a user's privclass is changed. All we do
            here is make sure the user is recorded as being in that
            privclass.
        """
        if not data['user'] in self.channel[data['ns']].member:
            return
        self.channel[data['ns']].member[data['user']]['pc'] = data['pc']
    
    def pkt_kicked(self, data):
        """ Received a kicked packet.
            
            This happens when the client is kicked from a channel.
            
            Here we automatically rejoin the channel if we are permitted
            to do so.
        """
        del self.channel[data['ns']]
        if self.flag.disconnecting or self.flag.quitting:
            return
        
        if 'r' in data.keys():
            if 'autokicked' in data['r'].lower() or 'not privileged' in data['r'].lower():
                if len(self.channel) > 0:
                    return
                self.handle_pkt(Packet('disconnect\ne=no joined channels\n\n'), time.time())
                return
        
        if self.flag.autorejoin:
            self.join(data['ns'])
    
    def pkt_disconnect(self, data):
        """ Received a disconnect packet from the server.
            
            Here we make sure the client does the right thing when a
            disconnect happens.
            
            This method is required for the client to work as expected.
        """
        self.flag.connected = False
        self.connection.disconnects+= 1
        self.channel = {}
        
        if self.flag.quitting:
            self.flag.close = True
            return
        
        if data['e'] == 'no joined channels' and len(self.autojoin) == 0:
            return
        
        if not self.flag.disconnecting:
            self.logger('>> Experiencing an unexpected disconnect.', showns=False)
            self.logger('>> Attempting to reconnect in a moment.', showns=False)
        
        time.sleep(1)
        self.nullflags()
        self.flag.reconnect = True
    
    # END PROTOCOL METHODS
    
    def parser(self):
        """ This method returns the :py:class:`Packet class
            <dAmnViper.parse.Packet>`.
        """
        return Packet
        
    def logger(self, msg, ns=None, showns=True, mute=False, pkt=None, ts=None):
        """ Write output to stdout. """
        if mute and not self.flag.debug:
            return
        
        if ns is None:
            ns = self.default_ns
        
        stamp = time.strftime('%H:%M:%S', time.localtime(ts))
        ns = ns + '|' if showns else ''
        
        try:
            sys.stdout.write('{0}|{1} {2}\n'.format(stamp, ns, msg))
        except:
            sys.stdout.write('{0}| >> I failed to display a message! Sorry.\n'.format(stamp))
        sys.stdout.flush()
        
    def new_logger(self, ns=None, showns=True, mute=False):
        """ Returns a wrapped logger method. """
        @wraps(self.logger)
        def wrapper(msg):
            return self.logger(msg, ns, showns, mute)
        return wrapper
    
    def get_write_pair(self, showns=False):
        """ Returns a pair of wrapped logger methods. """
        return (self.new_logger(showns=False), self.new_logger(showns=False, mute=(not self.flag.debug)))


class dAmnClient(Client):
    """ This class provides an easy to use API for connecting to dAmn, and for
        interacting with the server.
        
        This class inherits much of its behaviour from the :py:class:`Client
        class <dAmnViper.base.Client>`. This class has been made to cater more
        closely for dAmn servers. The Client class can be used to make clients
        which connect to llama based servers. In fact, with some slight
        modifications, this class could also connect to llama servers.
        
        Applications using this class should extend the class to add in
        some basic functionality so that it works as they expect. The
        class can be made to work properly with only minor modifications
        as shown here::
            
            from twisted.internet import reactor
            from dAmnViper.base import dAmnSock
            
            dAmn = dAmnClient()    
            
            dAmn.user.username = 'username'
            dAmn.user.token = 'authtoken'
            dAmn.autojoin = ['Botdom']
            
            dAmn.teardown = lambda: reactor.stop()
            
            dAmn.start()
            
            if dAmn.flag.connecting:
                reactor.run()
        
        The ``teardown`` callback needs to be defined or the application
        will hang when the connection to dAmn is lost.
        
        A way to remove the need to check the ``dAmnClient.flag.connecting``
        flag is to define your own ``on_connection_start`` method on the
        object, and using this to call ``reactor.run()``.
    """

    def __init__(self, *args, **kwargs):
        """ Initiate everything. """
        # Populate our attributes.
        self.populate_objects()
        self.agent = 'dAmnViper/{0} (Python) viper/base/dAmnClient/{1}.{2}'.format(
            self.platform.series, self.platform.version, self.platform.build)
        self.default_ns = '~Global'
        # Configure the class to connect to the dAmn server.
        self.CONST.SERVER = 'chat.deviantart.com'
        self.CONST.CLIENT = 'dAmnClient 0.3'
        self.CONST.PORT = 3900
        # Init?
        self.init(*args, **kwargs)
    
    def startedConnecting(self, connector):
        """ This method is called when we have started connecting. """
        self.logger('** Opening connection dAmn...', showns=False)
        self.flag.connecting = True
        self.on_connection_start(connector)
    
    def pkt_dAmnServer(self, data):
        """ Received a dAmnServer packet.
            
            This method calls the ``login`` method of this class and
            sets the ``shaking`` flag to ``False``.
            
            This method is essential for connecting to dAmn properly.
        """
        self.flag.shaking = False
        self.login()
    
# EOF



''' Example of a very basic dAmn bot!
    Created by photofroggy
'''

import sys
from twisted.internet import reactor

from dAmnViper.base import dAmnClient
from dAmnViper.examples.util import get_input

# Extend the dAmnViper.dAmnSock class to add some functionality.

class MyClient(dAmnClient):
    
    def init(self, username, authtoken, admin, trigger='!', autojoin=None, callbacks=None):
        """ Initialise the client.
            
            Override this method if you need to do anything when an
            instance of the object is created. Do not override __init__
            directly.
            
            This seems like bad practice but it does mean you don't have
            to remember to call the original __init__ method.
            
            In this example, we simply store the variables given to the
            constructor.
        """
        self.user.username = username
        self.user.token = authtoken
        self._admin = admin.lower()
        self.trigger = trigger
        self.autojoin = autojoin or ['Botdom']
        self.callbacks = callbacks or Commands()
    
    def on_connection_start(self, connector):
        """ This method is called when dAmnViper has started connecting.
            
            Use this method to start the reactor if needed. We can
            assume we have everything needed to connect, as this method
            is only called when dAmnViper actually succeeds in getting
            an authtoken.
            
            Technically you can achieve the same behaviour by using
            something similar to this outside of the class::
                
                dAmn.start()
                if dAmn.flag.connecting:
                    reactor.run()
            
            Using this method just makes things a little cleaner on the
            outside, but not really.
        """
        reactor.run()
    
    def teardown(self):
        """ Overriding this method is required to stop the application.
            This method is called by dAmnSock when the client has
            determined that it no longer needs to keep connected to
            dAmn, and has been fully disconnected from the server.
            
            If reactor.stop() is not called at this point, the program
            will hang indefinitely. Overriding onDisconnect is not
            recommended!
        """
        reactor.stop()
    
    def pkt_recv_msg(self, data):
        """ Basic command handling provided here.
            
            This method is called by dAmnSock when a 'recv_msg' packet
            is sent to the client by the server. By this point, dAmnSock
            should already have displayed a message in stdout reporting
            the packet.
            
            It is possible to add specific handling of certain packets
            by defining other pkt_* methods, but note that some are
            already defined, and need to be called in order for dAmnSock
            to function as expected.
            
            This example simply looks for the trigger character in the
            message, and then passes the message to a command handler
            where appropriate.
        """
        if data['message'][:len(self.trigger)] == self.trigger:
            data['message'] = data['message'][len(self.trigger):]
            data['args'] = data['message'].split(' ')
            self.callbacks.handle(
                data['args'][0].lower(), data, self)


class Commands(object):
    """ Just a simple object to hold any command callbacks. """
    
    def handle(self, cmd, data, client):
        getattr(self, 'cmd_{0}'.format(cmd),
            self.unknown_cmd)(data, client)
    
    def unknown_cmd(self, data, client):
        pass
    
    def cmd_about(self, data, dAmn):
        """Basic command callback."""
        dAmn.say(data['ns'], data['user']+': Basic dAmn Viper bot by photofroggy.')
    
    def cmd_quit(self, data, dAmn):
        """Quit command! Enter the admin name in place of 'admin'!"""
        if data['user'].lower() != dAmn._admin:
            return
        dAmn.say(data['ns'], data['user']+': Closing down!')
        dAmn.flag.quitting = True
        dAmn.disconnect()
    
    def cmd_refresh(self, data, dAmn):
        """Quit command! Enter the admin name in place of 'admin'!"""
        if data['user'].lower() != dAmn._admin:
            return
        dAmn.say(data['ns'], data['user']+': Refreshing connection!')
        dAmn.flag.disconnecting = True
        dAmn.disconnect()
        

def configure():
    """ Sort of explains itself.
        
        This is simply a method to request configuration detials from
        the user, via stdin.
    """
    sys.stdout.write('>> We need some details to be able to run the bot\n')
    
    obj = [get_input('>> Username: '),
        get_input('>> Authtoken: '),
        get_input('>> Admin: '),
        get_input('>> Trigger: '),
        [room.strip() for room in get_input('>> Autojoin: ', True).split(',')]
    ]
    
    while '' in obj[4]:
        obj[4].remove('')
    obj[4] = obj[4] or ['Botdom']
    
    return obj


if __name__ == '__main__':
    # Create a client
    sys.stdout.write('>> This is an example bot created with dAmn Viper.\n')
    
    try:
        dAmn = MyClient(*configure(), callbacks=Commands())
    except KeyboardInterrupt as e:
        sys.stdout.write('\n')
        sys.stdout.flush()
        sys.exit(0)
    
    sys.stdout.write('>> Starting the client...\n')
    sys.stdout.flush()
    
    # Start the dAmn client.
    dAmn.start()

# EOF



''' dAmnViper.net module
    Created by photofroggy.
    
    This module provides classes used to actually connect
    to dAmn using Twisted. The ConnectionFactory starts connections
    and handles disconnects. The IOProtocol handles basic IO operations
    on the connection, but delegates most of the processing to an
    instance of the dAmnSock class from the dAmnViper.base module.
'''

# Standard library
import time

# Twisted library imports
from twisted.internet import reactor
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import ClientFactory

# Viper stuff
from dAmnViper.parse import Packet


class ConnectionFactory(ClientFactory):
    """ Connection management.
        
        Instances of this class are used by the base :py:class:`Client class
        <dAmnViper.base.Client>` to connect to chat networks.
        
        The object returns instances of :py:class:`ChatProtocol
        <dAmnViper.net.ChatProtocol>` when a connection is made and
        handles disconnects and connection failures.
        
        This object should not be used directly by applications using
        dAmnViper.
    """
    
    stdout = None
    debug = None
    client = None

    def __init__(self, client, stdout=None, debug=None):
        """Initialise up in this motherfucka"""
        self.log = stdout
        self.debug = debug
        self.client = client
        
        if self.log is None:
            self.log = lambda m: None
        if self.debug is None:
            self.debug = lambda m: None
    
    def startedConnecting(self, connector):
        """ Called by twisted when we start to connect. """
        self.client.startedConnecting(connector)
    
    def create_protocol(self):
        return ChatProtocol(self, self.client, self.stdout, self.debug)
        
    def buildProtocol(self, addr):
        """ Called by twisted when a protocol is needed.
            
            Twisted calls this method when a connection can be
            established with the server. This method creates an instance
            of the :py:class:`ChatProtocol <dAmnViper.net.ChatProtocol>`
            class. This instance is given to the :py:class:`base Client
            <dAmnViper.base.Client>` object that the factory belongs
            to, and returned to twisted.
        """
        protocol = self.create_protocol()
        self.client.set_protocol(protocol)
        return protocol
    
    def clientConnectionLost(self, connector, reason):
        """ Called by twisted when a connection is lost.
            
            Displays a message notifying the connection loss and tells
            the base :py:class:`Client <dAmnViper.base.Client>` object
            that there is no longer a connection open.
        """
        
        self.client.connectionLost(connector, reason)
    
    def clientConnectionFailed(self, connector, reason):
        """ Called by twisted when we fail to connect properly.
            
            The behaviour of this method is similar to the
            ``clientConnectionLost`` method.
        """
        
        self.client.connectionFailed(connector, reason)


class ChatProtocol(Protocol):
    """ Protocol layer for the dAmn connection.
        
        Instances of this class are used to directly communicate with
        the connection to dAmn via twisted, and gives any data received
        to the :py:class:`Client <dAmnViper.base.Client>` instance
        being used for this connection.
        
        This object should not be used directly by applications using
        dAmnViper.
    """
    
    conn = None
    client = None
    log = None
    debug = None
    __buffer = None
    
    def __init__(self, conn, client, stdout=None, debug=None):
        """ Initialise the protocol. """
        # Store our objects
        self.conn = conn
        self.client = client
        self.log = stdout
        self.debug = debug
        self.__buffer = ''
        
        # Make sure we have callables for logging stuff.
        if self.log is None:
            self.log = lambda m: None
        
        if self.debug is None:
            self.debug = lambda m: None
    
    def connectionMade(self):
        """ Called by twisted when we have connected.
            
            This method simply calls the same method in the client object.
        """
        
        self.client.connectionMade()
    
    def dataReceived(self, data):
        """ Called by twisted when data is received.
            
            The data received is added to out buffer. If there are any
            full packets in the buffer, these packets are sent to the
            :py:class:`Client <dAmnViper.base.Client>` instance to
            be parsed properly.
            
            Any event handling relating to specific packet is done in the
            ``Client`` instance.
        """
        
        # Split on null.
        self.__buffer+= data
        raw = self.__buffer.split('\0')
        self.__buffer = raw.pop()
        
        for chunk in raw:
            packet = Packet(chunk)
            
            # If it's a ping packet, send a pong straight away!
            if packet.cmd == 'ping':
                self.send_packet('pong\n')
            
            # Let the client do whatever it needs to with the packet.
            self.client.handle_pkt(packet, time.time())
    
    def send_packet(self, data):
        """ A wrapper function for sending packets to the server.
            
            Here, a null character (``\\0``) is appended to the given
            data, and we return the number of characters we have tried
            send to the server.
        """
        data = '{0}\0'.format(data)
        self.transport.write(data)
        return len(data)


# EOF



''' Simply provides a cross-platform
    wrapper for getting input from the
    console.
'''

try:
    base = raw_input
except NameError as e:
    base = input

def get_input(prompt='>> ', empty=False):
    while True:
        data = base(prompt).strip()
        if not data and not empty:
            continue
        return data

# EOF



''' Example of a client using dAmn Viper's dAmnClient class!
    Created by photofroggy
'''

from twisted.internet import reactor

from dAmnViper.base import dAmnClient

if __name__ == '__main__':
    
    dAmn = dAmnClient()
    
    dAmn.user.username = 'username'
    dAmn.user.token = 'authtoken'
    dAmn.autojoin = ['Botdom']
    
    dAmn.teardown = lambda: reactor.stop()
    
    dAmn.start()
    
    if dAmn.flag.connecting:
        reactor.run()

# EOF



''' dAmnViper setup.
'''

from distutils.core import setup

setup(name='dAmnViper',
    version='3.56',
    description='dAmn connection library',
    author='photofroggy',
    author_email='froggywillneverdie@msn.com',
    url='https://github.com/photofroggy/dAmnViper',
    packages=[
        'dAmnViper',
        'dAmnViper.examples'
    ],
    platforms=['Any'],
    classifiers=[
        'Natural Language :: English',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
    ],
    long_description="""
==========
dAmn Viper
==========

**dAmn Viper** is a Python package created by photofroggy.

This package provides an API for connecting to and interacting with the dAmn
chats. That is, deviantART's messaging network chats.

This branch of dAmn Viper is only known to work with Python 2.7 and up due to
use of the twisted library.

-------------
Brief example
-------------

Creating a client quickly for dAmn is quite easy using dAmn Viper. Below
is one of the simplest examples of a client that simply connects, and
tries to stay connected::
    
    from twisted.internet import reactor
    from dAmnViper.base import dAmnClient
    
    dAmn = dAmnClient()
    
    dAmn.user.username = 'username'
    dAmn.user.token = 'authtoken'
    dAmn.autojoin = ['Botdom']
    
    dAmn.on_connection_start = lambda connector: reactor.start()
    dAmn.teardown = lambda: reactor.stop()
    
    dAmn.start()

That is all that is required! It is advised that you use the dAmnClient class
when making applications that connect to dAmn. Chat bots and full clients can
be made by extending the dAmnClient class to add functionality, as shown in the
examples provided.

------------
Dependencies
------------

A disadvantage of using twisted is the acquisition of dependencies. This means
that applications using dAmn Viper will depend on twisted as well as
dAmn Viper.

It is more of an issue for end users, as they will not want to spend
time installing multiple dependencies. This problem can, however, be
eleminated by creating installers for applications using dAmn Viper.

As such, this is somewhat a non-point, but it does mean an installer has to be
created to achieve easy setup for end users. Fortunately, installers are
something which users tend to be ok at using, so long as they aren't too
complicated. Having an installer which downloads and installs multiple
dependencies may complicate things too much. We'll see.

--------
Feedback
--------

Feedback on this branch of dAmn Viper would be very much appreciated.
Please send any feedback to my deviantART account via notes, or submit
something on github, I dunno.

Thanks for reading.

----------
DISCLAIMER
----------

Disclaimer::

		dAmn Viper is in no way affiliated with or endorsed by deviantART.com.
	This is not an official service of deviantART.com. This is an independent
	project created by Henry Rapley:
		<http://photofroggy.deviantart.com>
	
		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
	APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
	HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
	OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
	PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
	IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
	ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

"""
)

# EOF



''' dAmnViper.parse module
    Created by photofroggy.
    
    This module provides objects for parsing data going to and from
    the dAmn server. Tablumps and packets are handled by classes here,
    and a class is provided to generate output messages for stdout
    based on packets received from dAmn.
'''

import re

class Packet(object):
    """ Use this class to parse dAmn packets.
        
        This object processes given strings as dAmn packets, and stores
        information from the string in different object attributes. This
        makes it easier to work with packets in other parts of the API.
    """

    def __init__(self, data=None, sep='='):
        self.cmd, self.param, self.args, self.body, self.raw = None, None, {}, None, data
        
        if not data:
            return
        
        data = data.partition('\n\n')
        self.body = data[2] or None
        data = data[0].partition('\n')
        
        if not data[0]:
            return
        
        if not sep in data[0]:
            head = data[0].partition(' ')
            self.cmd = head[0] or None
            self.param = head[2] or None
            data = data[2].partition('\n')
        
        while data[0]:
            arg = data[0].partition(sep)
            data = data[2].partition('\n')
            if not arg[1]:
                continue
            self.args[arg[0]] = arg[2]
        
        # And that's the end of that chapter.

class Tablumps(object):
    """ dAmn tablumps parser.
        
        dAmn sends certain information formatted in a specific manner.
        Links, images, thumbs, and other forms of data are formatted
        in strings where the different attributes of these values are
        separated by tab characters (``\\t``), and usually begin with an
        ampersand.
        
        We refer to these items as "tablumps" because of the tab
        characters being used as delimeters. The job of this class is to
        replace tablumps with readable strings, or to extract the data
        given in the tablumps.
    """
    
    expressions = None
    replace = None
    titles = None
    subs = None
    
    def __init__(self):
        """Populate the expressions and replaces used when parsing tablumps."""
        if self.expressions is not None:
            return
        # Regular expression objects used to find any complicated tablumps.
        self.expressions = [
            re.compile("&avatar\t([a-zA-Z0-9-]+)\t([0-9]+)\t"),
            re.compile("&dev\t(.)\t([a-zA-Z0-9-]+)\t"),
            re.compile("&emote\t([^\t]+)\t([0-9]+)\t([0-9]+)\t(.*?)\t([a-z0-9./=-]+)\t"),
            re.compile("&a\t([^\t]+)\t([^\t]*)\t"),
            re.compile("&link\t([^\t]+)\t&\t"),
            re.compile("&link\t([^\t]+)\t([^\t]+)\t&\t"),
            re.compile("&acro\t([^\t]+)\t(.*)&\/acro\t"),
            re.compile("&abbr\t([^\t]+)\t(.*)&\/abbr\t"),
            re.compile("&thumb\t(?P<ID>[0-9]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t"),
            re.compile("&img\t([^\t]+)\t([^\t]*)\t([^\t]*)\t"),
            re.compile("&iframe\t([^\t]+)\t([0-9%]*)\t([0-9%]*)\t&\/iframe\t"),
        ]
        self.titles = ('avatar', 'dev', 'emote', 'a', 'link', 'link', 'acronym', 'abbr', 'thumb', 'img', 'iframe')
        # Regular expression objects used to find and replace complicated tablumps.
        self.subs = [
            (re.compile("&avatar\t([a-zA-Z0-9-]+)\t([0-9]+)\t"), ":icon\\1:"),
            (re.compile("&dev\t(.)\t([a-zA-Z0-9-]+)\t"), ":dev\\2:"),
            (re.compile("&emote\t([^\t]+)\t([0-9]+)\t([0-9]+)\t(.*?)\t([a-z0-9./=-]+)\t"), "\\1"),
            (re.compile("&a\t([^\t]+)\t([^\t]*)\t"), "<a href=\"\\1\" title=\"\\2\">"),
            (re.compile("&link\t([^\t]+)\t&\t"), "\\1"),
            (re.compile("&link\t([^\t]+)\t([^\t]+)\t&\t"), "\\1 (\\2)"),
            (re.compile("&acro\t([^\t]+)\t"), "<acronym title=\"\\1\">"),
            (re.compile("&abbr\t([^\t]+)\t"), "<abbr title=\"\\1\">"),
            (re.compile("&thumb\t([0-9]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t"), ":thumb\\1:"),
            (re.compile("&img\t([^\t]+)\t([^\t]*)\t([^\t]*)\t"), "<img src=\"\\1\" alt=\"\\2\" title=\"\\3\" />"),
            (re.compile("&iframe\t([^\t]+)\t([0-9%]*)\t([0-9%]*)\t&\/iframe\t"), "<iframe src=\"\\1\" width=\"\\2\" height=\"\\3\" />"),
            (re.compile("<([^>]+) (width|height|title|alt)=\"\"([^>]*?)>"), "<\\1\\3>"),
        ]
        # Search and replace pairs used to parse simple HTML tags.
        self.replace = [
            ("&b\t", "<b>"),
            ("&/b\t", "</b>"),
            ("&i\t", "<i>"),
            ("&/i\t", "</i>"),
            ("&u\t", "<u>"),
            ("&/u\t", "</u>"),
            ("&s\t", "<s>"),
            ("&/s\t", "</s>"),
            ("&sup\t", "<sup>"),
            ("&/sup\t", "</sup>"),
            ("&sub\t", "<sub>"),
            ("&/sub\t", "</sub>"),
            ("&code\t", "<code>"),
            ("&/code\t", "</code>"),
            ("&p\t", "<p>"),
            ("&/p\t", "</p>"),
            ("&ul\t", "<ul>"),
            ("&/ul\t", "</ul>"),
            ("&ol\t", "<ol>"),
            ("&/ol\t", "</ol>"),
            ("&li\t", "<li>"),
            ("&/li\t", "</li>"),
            ("&bcode\t", "<bcode>"),
            ("&/bcode\t", "</bcode>"),
            ("&br\t", "\n"),
            ("&/a\t", "</a>"),
            ("&/acro\t", "</acronym>"),
            ("&/abbr\t", "</abbr>"),
        ]
    
    def parse(self, data):
        """ Parse any dAmn Tablumps found in our input data.
            
            This method will simply return a string with the tablumps
            parsed into readable formats.
        """
        try:
            for lump, repl in self.replace:
                data = data.replace(lump, repl)
            for expression, repl in self.subs:
                data = expression.sub(repl, data)
        except Exception:
            pass
        return data
    
    def capture(self, text):
        """ Return any dAmn Tablumps found in our input data.
            
            Rather than parsing the tablumps, this method returns the
            data given by tablumps. This only works for tablumps where
            a regular expression is used for parsing.
        """
        lumps = {}
        for key, expression in enumerate(self.expressions):
            cc = expression.findall(text)
            if not cc:
                continue
            lumps[self.titles[key]] = cc
        return lumps

class ProtocolParser(object):
    """ Parser for processing the dAmn protocol.
        
        Use methods of this class to process dAmn information. You can
        customise how different packets are handled by adding items to
        protocol.maps and protocol.messages.
    """
    maps = {
        'dAmnServer': ['version'],
        'login': ['username', ['e'], 'data'],
        'join': ['ns', ['e'] ],
        'part': ['ns', ['e', '*r'] ],
        'property': ['ns', ['p', 'by', 'ts'], '*value' ],
        'recv_msg': [None, [('from', 'user')], '*message'],
        'recv_action': [None, [('from', 'user')], '*message'],
        'recv_join': ['user', None, '*info'],
        'recv_part': ['user', ['r']],
        'recv_privchg': ['user', ['by', 'pc']],
        'recv_kicked': ['user', ['by'], '*r'],
        'recv_admin_create': [None, ['p', ('by', 'user'), ('name', 'pc'), 'privs']],
        'recv_admin_update': [None, ['p', ('by', 'user'), ('name', 'pc'), 'privs']],
        'recv_admin_rename': [None, ['p', ('by', 'user'), 'prev', ('name', 'pc')]],
        'recv_admin_move': [None, ['p', ('by', 'user'), 'prev', ('name', 'pc'), ('n', 'affected')]],
        'recv_admin_remove': [None, ['p', ('by', 'user'), ('name', 'pc'), ('n', 'affected')]],
        'recv_admin_show': [None, ['p'], 'info'],
        'recv_admin_showverbose': [None, ['p'], 'info'],
        'recv_admin_privclass': [None, ['p', 'e'], 'command'],
        'kicked': ['ns', [('by', 'user')], '*r'],
        'ping': [],
        'disconnect': [None, ['e']],
        'send': ['ns', ['e']],
        'kick': ['ns', [('u', 'user'), 'e']],
        'get': ['ns', ['p', 'e']],
        'set': ['ns', ['p', 'e']],
        'kill': ['ns', ['e']],
        'unknown': [None, None, None, None, 'packet'],
    }
    names = [('recv', 'admin')]
    messages = {
        'dAmnServer': ('** Connected to dAmn Server {0}.', False),
        'login': ('** Login as {0}: {1}.', False),
        'join': ('** Join {ns}: "{1}".', False),
        'part': ('** Part {ns}: "{1}". [{2}]', False),
        'property': ('** Got {1} for {ns}.', False),
        'recv_msg': ('<{1}> {2}',),
        'recv_action': ('* {1} {2}',),
        'recv_join': ('** {1} has joined.',),
        'recv_part': ('** {1} has left. [{2}]',),
        'recv_privchg': ('** {1} has been made a member of {3} by {2} *',),
        'recv_kicked': ('** {1} has been kicked by {2} * {3}',),
        'recv_admin_create': ('** Privilege class {3} has been created by {2} with: {4}',),
        'recv_admin_update': ('** Privilege class {3} has been updated by {2} with: {4}',),
        'recv_admin_rename': ('** Privilege class {4} has been renamed to {3} by {2}',),
        'recv_admin_move': ('** All members of {3} have been moved to {4} by {2} -- {5} affected user(s)',),
        'recv_admin_remove': ('** Privilege class {3} has been removed by {2} -- {4} affected user(s)',),
        'recv_admin_show': None,
        'recv_admin_showverbose': None,
        'recv_admin_privclass': ('** Admin command "{2}" failed: {1}',),
        'kicked': ('** You have been kicked by {1} * {2}',),
        'ping': ('** Ping...', False, True),
        'disconnect': ('** You have been disconnected * {0}', False),
        'send': ('** Send error: {1}',),
        'kick': ('** Could not kick {1}: {2}',),
        'get': ('** Could not get {1}info for {ns}: {2}', False),
        'set': ('** Could not set {1}: {2}',),
        'kill': ('** Kill error: {1}',),
        'unknown': ('** Received unknown packet in {0}: {1}', False),
    }
    
    tablumps = Tablumps
    
    def __init__(self):
        self.tablumps = self.tablumps()
    
    def evt_namespace(self, pkt):
        # Work on the namespace of an event!
        namespace = pkt.cmd
        for conditions in self.names:
            if conditions[0] == namespace:
                subline = pkt.body.split('\n')[0]
                subline = subline.split(' ')
                if not subline or not subline[0]:
                    break
                namespace = '_'.join([namespace, subline[0]])
                if len(conditions) > 1 and len(subline) > 1:
                    if conditions[1] == subline[0]:
                        namespace = '_'.join([namespace, subline[1]])
                break
        return namespace if namespace in self.maps else 'unknown'
    
    def mapper(self, pkt):
        """ Map packets to a usable data structure.
            
            Return data mapped according to the conditions defined in
            the ``maps`` attribute.
        """
        store = {'args': {}, 'rules': [], 'event': self.evt_namespace(pkt)}
        map = self.maps[store['event']]
        return getattr(self, 'gen_'+pkt.cmd, self.sort)(store, pkt, map)
        
    def sort(self, store, data, map):
        """ Sort data according to the conditions in the given map. """
        for i, cond in enumerate(map):
            if not cond: continue
            if i is 0:
                ptab = cond[0] == '*'
                if ptab: cond = cond[1:]
                val = (data.param if not ptab else self.tablumps.parse(data.param)) if data.param else ''
                store['rules'].append((cond, val))
                store['args'][cond] = val
            if i is 1:
                for item in cond:
                    if not item:
                        continue
                    argn, name = item if isinstance(item, tuple) else (item, item)
                    ptab = argn[0] == '*'
                    if ptab and argn == name: name = argn[1:]
                    val = '' if not argn in data.args.keys() else data.args[argn]
                    if ptab: val = self.tablumps.parse(val)
                    store['rules'].append((name, val))
                    store['args'][name] = val
            if i is 2:
                ptab = cond[0] == '*'
                if ptab: cond = cond[1:]
                val = (data.body if not ptab else self.tablumps.parse(data.body)) if data.body else ''
                store['rules'].append((cond, val))
                store['args'][cond] = val
            if i is 3:
                store = self.sort(store, Packet(data.body), cond)
            if i is 4:
                val = data.raw if data.raw else ''
                store['rules'].append((cond, val))
                store['args'][cond] = val
        return store
    
    def logger(self, event, data, ns, pkt):
        """ Return a log_list (message, channel[, bool(showns)[, bool(mute)]]).
            
            The message returned is based on the templates defined in
            the ``messages`` attribute of this class.
        """
        sequence = ['', ns, True, False, pkt]
        
        if not event in self.messages.keys():
            return None
        
        if hasattr(self, 'log_'+event):
            return getattr(self, 'log_'+event)(event, data, ns, pkt)
        
        msgtpl = self.messages[event]
        
        if msgtpl is None or len(msgtpl) == 0:
            return None
        
        data = [(item[1] if bool(item[1]) else '') for item in data]
        disp = (msgtpl[0].replace('{ns}', ns)).format(*data)
        
        if disp[-3:] == ' []':
            disp = disp[:-3]
        
        sequence[0] = disp
        msgtpl = [] if len(msgtpl) == 1 else msgtpl[1:]
        i = 2
        
        for item in msgtpl:
            sequence[i] = item
            i+=  1
        
        return sequence
    
    def gen_recv(self, store, data, map):
        # Generic recv packet mapping operations.
        store['args']['ns'] = data.param
        store['rules'].append(('ns', data.param))
        store = self.sort(store, Packet(data.body), map)
        if store['event'] in ('recv_msg', 'recv_action'):
            store['args']['raw'] = data.raw
            store['rules'].append(('raw', data.raw))
        return store

# EOF
